from plow.core.logtail import LogTail
from event_api_client.core import EventAPI, Event
from plow.core.time_parse import utcnow, pacific_now
from customer_info.ci import CustomerInfo, Person
from plow.core.helpers import MetricState, EventState, Timer
from pygrok import Grok
from abc import ABC
import traceback
import sys
import os
import arrow


class Scrape(ABC):
    def __init__(self, logger, app_config=None):
        # WARNING : DO NOT PUT THINGS IN INIT THAT WILL CONTINUE TO GROW IN SIZE
        # scheduler calls init once and so these objects exist forever.
        self.scrape_type = type(self).__name__.lower()
        self.log = logger
        if hasattr(self.log, '_meta'):
            self.log._meta.namespace = "plow:scrape:%s" % (self.scrape_type)

        # TODO this will be the one and only state object
        self.app_config = app_config
        self.event_api = EventAPI(
            api_url=app_config['event_api_url'], api_token=app_config['event_api_token'])

    def disable_metrics(self):
        metrics = self.app_config.get('disable_metrics')
        if metrics:
            return metrics.split(",")
        else:
            return []

    def grok(self, pattern):
        try:
            return Grok(pattern, fullmatch=False)
        except Exception as err:
            self.log.error(
                "failed to create grok object %s: %s" % (pattern, err))
            traceback.print_exc(file=sys.stdout)
            return

    def matches(self, pattern=None):
        matches = []
        grok = self.grok(pattern)
        for line in self.lines:
            try:
                # TODO this needs a timeout cause it causes it can hang on some
                # failed matches
                match = grok.match(line.rstrip())
                if match:
                    matches.append(match)
            except Exception as err:
                continue

        return matches

    def matches_by_field(self, field, matches):
        data = {}
        for m in matches:
            data.setdefault(m[field], []).append(m)
        return data

    def metrics(self):
        return {}

    def get_pattern(self, name=None):
        parsers = self.parsers()
        if name in parsers:
            return parsers[name]
        else:
            return parsers['default']

    def parsers(self):
        return {'default': ''}

    def run_metrics(self):
        for name in self.metrics():
            if name in self.disable_metrics():
                continue
            self.log.debug("starting %s" % name)
            with Timer() as t:
                pattern = self.get_pattern(name)
                if pattern:
                    matches = self.matches(pattern=pattern)
                    getattr(self, name)(matches)

            self.metrics_state.gauge(
                'plow_scrape_time',
                tags=['scrape'],
                tag_values=[name],
                value=t.seconds
            )

            self.log.debug("finished %s in %s seconds" %
                           (name, t.seconds))

    def do_run(self, **kwargs):
        self.log.info("started scrape %s" % self.scrape_type)
        self.events = []
        self.lines = []
        # need fresh customer info every run
        self.customer_info = CustomerInfo()

        self.metrics_state = MetricState(
            name='scrape-' + self.scrape_type,
            path=self.app_config['metrics_path'],
        )
        self.state_events = EventState(
            name='events-scrape-' + self.scrape_type,
            path=self.app_config['state_path'],
            rotate=60
        )

        self.state_events.rotate()

        with Timer() as t:
            self.run(**kwargs)

        self.log.info("completed scrape %s in %s seconds" % (
            self.scrape_type, t.seconds))

        self.metrics_state.gauge(
            'plow_scrape_total_time',
            tags=['scrape'],
            tag_values=[self.scrape_type],
            value=t.seconds
        )

        self.metrics_state.gauge(
            'plow_scrape_lines',
            tags=['scrape'],
            tag_values=[self.scrape_type],
            value=len(self.lines)
        )

        self.metrics_state.dump(debug=self.app_config.get('debug'))
        self.log.info("sending events")
        self.send_events()

    def run(self, logs=None, fast_scrape=False, **kwargs):
        if not logs:
            self.log.error("no logs")
            return

        log_tail = LogTail()
        self.lines = []
        for log in log_tail.tail(logs):
            lines = log.readlines()
            if fast_scrape:
                if len(lines) > 5000:
                    self.log.debug("fast_scrape.. reading from EOF")
                    log.read_from_end = True
                    lines += log.readlines()

            self.lines += lines

        self.run_metrics()

    def add_event(self, **event):
        try:
            person = self.customer_info.get_person(event['person'])
            if person and person.should_send_events():
                event_obj = Event(event)
                self.events.append(event_obj)
        except Exception as e:
            self.log.error("failed to add event %s" % e)

    def send_events(self):
        expire_values = {
            'once_person_object': 86400 * 30,
            'once_person': 86400 * 30
        }
        for e in (sorted(self.events, key=lambda i: i.stamp)):
            expire = 86400
            if e.event_type in expire_values:
                expire = expire_values[e.event_type]

            # if this is a first type of event, was their account made recently?
            if not self.state_events.get(e.event_key_string()):
                self.log.debug("sending event: %s" % e.dump())
                resp = self.event_api.send(e)
                if resp:
                    self.state_events.set(
                        resp.event_key_string(), True)
