import re
import os
import json
import arrow
from ipaddress import IPv4Address
import statistics
from state_manager.seen_ip import SeenIPState
from plow.core.scrape import Scrape
from crawler_user_agents.parse import CrawlerAgent
from plow.core.rdap import NetworkOwner
from plow.core.time_parse import apache_stamp_to_iso, apache_stamp_to_arrow, pacific_now
from plow.core.helpers import geo_lookup_ip


class Apache(Scrape):
    def __init__(self, logger=None, app_config=None):
        super().__init__(logger=logger, app_config=app_config)
        self.crawler_agent = CrawlerAgent()
        self.network_owner = NetworkOwner()

    def get_seen_ip_state(self):
        seen_ip_state = SeenIPState(
            path=self.app_config['state_path'], rotate=1)
        seen_ip_state.initdb()
        seen_ip_state.rotate()
        return seen_ip_state

    def metrics(self):
        return {'request_uniq', 'request_time', 'request', 'request_origin'}

    def parsers(self):
        return {
            'default': '%{WORD:person} %{WORD:user} %{HOSTNAME:domain} \[%{HTTPDATE:timestamp}\] %{IPORHOST:clientip} "(?:%{WORD:method} %{NOTSPACE:request}(?: HTTP/%{NUMBER:httpversion})?|%{DATA:rawrequest})" %{NUMBER:response} %{QS:agent} "%{DATA:req_file}" %{NUMBER:req_time}',
        }

    def _generic_response_code(self, code):
        return "%sXX" % code[:1]

    def match_mean(self, name, matches):
        if not len(matches):
            return '0.0'
        numbers = [int(m[name]) for m in matches]
        mean = statistics.mean(numbers)
        return mean

    def matches_by_type(self, matches):
        bots = []
        users = []
        errors = []
        for m in matches:
            if self.is_bot(m):
                bots.append(m)
            else:
                users.append(m)
            if m['response'].startswith('5'):
                errors.append(m)
        return (bots, users, errors)

    def matches_by_bot_type(self, matches):
        data = {}
        for m in matches:
            bot_ua = self.is_bot(m)
            if bot_ua:
                # default is not happy here for some reason
                if bot_ua not in data:
                    data[bot_ua] = []
                data[bot_ua].append(m)
        return data

    def is_blocked(self, match):
        if match['response'] == '415':
            return True

    # returns a bot name or nothing if its not a bot
    def is_bot(self, match):
        ca = self.crawler_agent.match(match['agent'])
        if ca:
            return ca.name
        # is it from a private addr
        if ':' not in match['clientip'] and IPv4Address(match['clientip']).is_private:
            return 'ndn_net_private'
        # is it from one of our ip's
        if self.network_owner.get(match['clientip']) == 'NDN':
            return 'ndn_net'

    def request_uniq(self, matches):
        seen_ip_state = self.get_seen_ip_state()
        bots, not_bots, errors = self.matches_by_type(matches)
        domain_seenip = dict()
        for m in not_bots:
            domain = m['domain']
            clientip = m['clientip']
            if domain not in domain_seenip:
                domain_seenip[domain] = []
            if clientip not in domain_seenip[domain]:
                domain_seenip[domain].append(clientip)

        total_uniq = 0
        for d in domain_seenip:
            person = self.customer_info.person_by_domain(d)
            domain = self.customer_info.get_domain(d)
            seen_ip_state.add_ips(domain.domain, ips=domain_seenip[d])
            count_value = seen_ip_state.count_ips(domain.domain)
            self.metrics_state.gauge(
                'requests_uniq_by_domain',
                tags=['person', 'domain'],
                tag_values=[person.person, domain.domain],
                value=count_value
            )
            total_uniq += count_value

        self.metrics_state.gauge(
            'requests_uniq_total',
            tags=['type'],
            tag_values=['agg'],
            value=total_uniq
        )

    def request_time(self, matches):
        php_matches = [m for m in matches if m['req_file'].endswith('.php')]
        php_cgi_matches = [m for m in matches if re.match(
            "(.*)/php\d+.cgi$", m['req_file'])]
        not_php_matches = [
            m for m in matches if not m['req_file'].endswith('.php') and not re.match("(.*)/php\d+.cgi$", m['req_file'])
        ]

        self.request_time_gauge_filtered('php', php_matches)
        self.request_time_gauge_filtered('php_cgi', php_cgi_matches)
        self.request_time_gauge_filtered('not_php', not_php_matches)

    def request_time_gauge_filtered(self, name, matches):
        mean = self.match_mean('req_time', matches)
        self.metrics_state.gauge(
            'request_time_all',
            tags=['type'],
            tag_values=[name],
            value=mean,
        )

        for n in range(1, 5):
            n = str(n)
            n_filter = [m for m in matches if m['response'].startswith(n)]
            n_mean = self.match_mean('req_time', n_filter)
            generic_code = self._generic_response_code(n)
            self.metrics_state.gauge(
                'request_time_by_response_all',
                tags=['code', 'type'],
                tag_values=[generic_code, name],
                value=n_mean,
            )

            matches_by_domain = self.matches_by_field('domain', n_filter)
            for domain, domain_matches in matches_by_domain.items():
                self.metrics_state.gauge(
                    'request_time_response_by_domain',
                    tags=['code', 'type', 'domain'],
                    tag_values=[generic_code, name, domain],
                    value=self.match_mean('req_time', domain_matches)
                )

    def request(self, matches):
        """
            creates request couters for bots,users and errors
            bot = user agent matching a crawler
            user = anything not matching a crawler

        """

        bots, not_bots, errors = self.matches_by_type(matches)
        self.metrics_state.counter(
            'requests_all',
            tags=['type'],
            tag_values=['bot'],
            step=len(bots)
        )
        self.metrics_state.counter(
            'requests_all',
            tags=['type'],
            tag_values=['user'],
            step=len(not_bots)
        )
        self.metrics_state.counter(
            'requests_all',
            tags=['type'],
            tag_values=['error'],
            step=len(errors)
        )
        for method, method_matches in self.matches_by_field('method', matches).items():
            self.metrics_state.counter(
                'request_method_all',
                tags=['method'],
                tag_values=[method],
                step=len(method_matches)
            )

        for code, code_matches in self.matches_by_field('response', matches).items():
            self.metrics_state.counter(
                'request_response_all',
                tags=['code'],
                tag_values=[code],
                step=len(code_matches)
            )
            self.metrics_state.counter(
                'request_response_generic_all',
                tags=['code'],
                tag_values=[self._generic_response_code(code)],
                step=len(code_matches)
            )

        for bot_ua, bot_matches in self.matches_by_bot_type(bots).items():
            self.metrics_state.counter(
                'useragent_bot_all',
                tags=['browser_family'],
                tag_values=[bot_ua],
                step=len(bot_matches)
            )
            if bot_ua == 'WordPress':
                wp_cron_matches = [m for m in bot_matches if m['request'] is not None and re.match(
                    '(.*)wp-cron.php(.*)', m['request'])]
                self.metrics_state.counter(
                    'wordpress_request_all',
                    tags=['type'],
                    tag_values=['wp-cron'],
                    step=len(wp_cron_matches)
                )
            if bot_ua == 'Jetpack':
                self.metrics_state.counter(
                    'wordpress_request_all',
                    tags=['type'],
                    tag_values=['jetpack'],
                    step=len(bot_matches)
                )

        xmlrpc_matches = [m for m in matches if m['request'] is not None and re.match(
            '(.*)wp-cron.php(.*)', m['request'])]
        self.metrics_state.counter(
            'wordpress_request_all',
            tags=['type'],
            tag_values=['xmlrpc'],
            step=len(xmlrpc_matches)
        )

        # TODO move domain metrics to their own counter
        for domain, domain_matches in self.matches_by_field('domain', matches).items():
            d_bots, d_not_bots, d_errors = self.matches_by_type(domain_matches)
            person = self.customer_info.person_by_domain(domain)
            self.metrics_state.counter(
                'requests_by_domain',
                tags=['person', 'domain', 'type'],
                tag_values=[person.person, domain, 'bot'],
                step=len(d_bots)
            )
            self.metrics_state.counter(
                'requests_by_domain',
                tags=['person', 'domain', 'type'],
                tag_values=[person.person, domain, 'user'],
                step=len(d_not_bots)
            )

            self.metrics_state.counter(
                'requests_by_domain',
                tags=['person', 'domain', 'type'],
                tag_values=[person.person, domain, 'error'],
                step=len(d_errors)
            )

    def domain_metrics_gauge(self, name, matches):
        # TODO this is not working yet
        return
        data = {}
        for m in matches:
            data.setdefault(m['domain'], []).append(m)

        results = []
        results_by_mysql_host = []
        for domain in data:
            d = self.customer_info.get_domain(domain)
            mean = self.match_mean('req_time', data[domain])
            result = {
                'domain': domain,
                'req_time_mean': mean
            }
            if hasattr(d, 'wordpress'):
                if 'mysql_host' in d.wordpress:
                    result['mysql_host'] = d.wordpress['mysql_host']

            results.append(result)

        mean = self.match_mean('req_time_mean', results)

    def request_origin(self, matches):
        network_owner_counts = {}
        country_counts = {}
        for ip, ip_matches in self.matches_by_field('clientip', matches).items():
            no = self.network_owner.get(ip)
            if no:
                if no in network_owner_counts:
                    network_owner_counts[no] += len(ip_matches)
                else:
                    network_owner_counts[no] = len(ip_matches)
            c = geo_lookup_ip(ip)
            if c:
                if c in country_counts:
                    country_counts[c] += len(ip_matches)
                else:
                    country_counts[c] = len(ip_matches)

        for network_owner in network_owner_counts:
            self.metrics_state.counter(
                'requests_by_origin',
                tags=['origin'],
                tag_values=[network_owner],
                step=network_owner_counts[network_owner]
            )
        for cc in country_counts:
            self.metrics_state.counter(
                'requests_by_country',
                tags=['country'],
                tag_values=[cc],
                step=country_counts[cc]
            )
